from spb import MB, plot_list
from spb.series import BaseSeries
from spb.defaults import TWO_D_B
from spb.backends.matplotlib.renderers.renderer import MatplotlibRenderer
from numpy import array
from .stats import frequency, group_percentile, group_mean
from matplotlib.cbook import boxplot_stats

# Based on
# https://sympy-plot-backends.readthedocs.io/en/latest/tutorials/tut-6.html
# https://github.com/Davide-sd/sympy-plot-backends/blob/master/spb/series.py#L1238


class BoxplotSeries(BaseSeries):
    """Representation for a boxplot of ungrouped values."""

    def __init__(self, bxp_dict, **kwargs):
        super().__init__(**kwargs)
        self.bxp_dict = bxp_dict

    def get_data(self):
        return self.bxp_dict

    def __str__(self):
        return f"Boxplot"


class BarplotSeries(BaseSeries):
    def __init__(self, x, y, xticks, label="", width=0.8, **kwargs):
        super().__init__(**kwargs)

        self.x = x
        self.y = y
        self.xticks = xticks

        self.rendering_kw.setdefault("edgecolor", "black")
        self.rendering_kw.setdefault("width", width)
        if label != "":
            self.rendering_kw.setdefault("label", [label] + ["_"] * (len(x) - 1))
            self.show_in_legend = True
            self.legend = True
        else:
            self.show_in_legend = False

    def get_data(self):
        return self.x, self.y, self.xticks

    def __str__(self):
        return f"Barplot"


def draw_box(renderer, data):
    p, s = renderer.plot, renderer.series
    handle = p._ax.bxp(data, **s.rendering_kw)
    return handle


def draw_bar(renderer, data):
    p, s = renderer.plot, renderer.series
    x, y, xticks = data
    p.axis_center = None
    handle = p._ax.bar(x, y, **s.rendering_kw)
    p._ax.set_xticks(xticks)
    if s.show_in_legend:
        p.legend = True
    return handle


def update(_renderer, _data, _handle):
    raise NotImplementedError


class BoxplotRenderer(MatplotlibRenderer):
    draw_update_map = {draw_box: update}


class BarplotRenderer(MatplotlibRenderer):
    draw_update_map = {draw_bar: update}


MB.renderers_map.update({BoxplotSeries: BoxplotRenderer})
MB.renderers_map.update({BarplotSeries: BarplotRenderer})


def boxplot(data, groups=[], labels=[], **kwargs):
    """
    Visualize boxplot(s) of ungrouped observations. Typical usage examples:

    - Plotting a dataset:
        `boxplot(data,labels=["data"],**kw)`

    - Plotting multiple datasets:
        `boxplot([data1,data2], **kw)`

    Parameters
    ==========
    groups : list, optional
        List of edge points in each interval. Each element can consist of a single or two edge values.
        Can either be a single list for all lists of observations or individual lists for each.

    labels : list, optional
        Labels to be shown for each dataset
    """
    show = kwargs.get("show", True)
    backend = kwargs.get("backend", TWO_D_B)

    if backend != MB:
        raise NotImplemented

    stats = []
    if groups != []:
        if not hasattr(data[0], "__iter__"):
            data = [data]
        for d in data:
            p = group_percentile(d, groups, [0, 25, 50, 75, 100])
            stats.append(
                {
                    "mean": group_mean(d, groups),
                    "med": p[2],
                    "q1": p[1],
                    "q3": p[3],
                    "iqr": p[3] - p[1],
                    "cilo": 0,
                    "cihi": 0,
                    "whislo": p[0],
                    "whishi": p[4],
                    "fliers": [],
                }
            )

    else:
        stats = boxplot_stats(data, whis=(0, 100))

    if labels != []:
        for i, l in enumerate(labels):
            stats[i]["label"] = l

    p = backend(BoxplotSeries(stats), **kwargs)

    if show:
        p.show()
    return p


def smallest_gap(data: list):
    data.sort()
    gaps = [x - y for x, y in zip(data[1:], data[:-1])]
    return min(gaps)


def plot_bars(data, labels=[""], **kwargs):
    """
    Visualize bar plot(s) of ungrouped observations. Typical usage examples:

    - Plotting a dataset:
        `plot_bars(data,labels="data",**kw)`

    - Plotting multiple datasets:
        `plot_bars([data1,data2], **kw)`

    Parameters
    ==========
    labels : list, optional
        Labels to be shown for each dataset

    width : float or list, optional
        Width of bars
    """
    show = kwargs.get("show", True)
    backend = kwargs.get("backend", TWO_D_B)

    if backend != MB:
        raise NotImplemented

    kwargs.setdefault("ylabel", "Frekvens %")

    if hasattr(data[0], "__iter__"):
        series = []
        global_min = data[0][0]
        global_x = set()

        for d in data:
            f = frequency(d, show=False)
            global_x.update(f.observation)
        global_x = list(global_x)
        gap = smallest_gap(global_x)
        width = kwargs.get("width", gap / (len(data) + 1))
        start = -gap / 2 + width

        if labels != [""]:
            assert len(labels) == len(data)
        else:
            labels = [""] * len(data)

        for i, d in enumerate(data):
            f = frequency(d, show=False)
            offset = start + i * width
            x = (array(f.observation) + offset).tolist()
            xticks = kwargs.get("xticks", global_x)
            series.append(BarplotSeries(x, f.frekvens, xticks, labels[i], width=width))
            if min(d) < global_min:
                global_min = min(d)

        # kwargs.setdefault("axis_center", (global_min - gap, 0))
        # kwargs.setdefault("axis_center", "None")
        p = backend(*series, **kwargs)
    else:
        if hasattr(labels, "__iter__"):
            labels = labels[0]
        f = frequency(data, show=False)
        x = f.observation
        gap = smallest_gap(x)
        width = kwargs.get("width", 0.8 * gap)
        # kwargs.setdefault("axis_center", (min(data) - 0.8*width, 0))
        p = backend(BarplotSeries(x, f.frekvens, x, labels, width), **kwargs)
    # p.ax.spines["left"].set_position(("data", 0))
    if show:
        p.show()
    return p


def plot_sum(data, groups=[], labels=[], **kwargs):
    """
    Visualize Cumulative Distribution Function (CDF) of discrete observations. Typical usage examples:

    - Plotting a ungrouped dataset:
        `plot_sum(data,labels="data",**kw)`

    - Plotting multiple grouped datasets:
        `plot_sum([data1,data2],groups=[1, 2, 3, 4, 5], **kw)`

    Parameters
    ==========
    groups : list, optional
        List of edge points in each interval. Each element can consist of a single or two edge values.
        Can either be a single list for all lists of observations or individual lists for each.

    labels : list, optional
        Labels to be shown for each dataset
    """
    kwargs.setdefault("ylabel", "Kumuleret Frekvens %")

    datasets = []
    if not hasattr(data[0], "__iter__"):
        data = [data]

    if groups != []:
        for d in data:
            f = frequency(d, groups, show=False)
            xx, yy = [f.observation[0][0]], [0]
            for i, kf in enumerate(f.kumuleret_frekvens):
                xx.append(f.observation[i][1])
                yy.append(kf)
            datasets.append((xx, yy))

    else:
        for d in data:
            f = frequency(d, show=False)
            xx, yy = [f.observation[0]], [0]
            for i, kf in enumerate(f.kumuleret_frekvens):
                if i > 0:
                    xx.append(f.observation[i])
                    yy.append(f.kumuleret_frekvens[i - 1])
                xx.append(f.observation[i])
                yy.append(kf)
            datasets.append((xx, yy))

    return plot_list(*datasets, label=labels, **kwargs)


def plot_hist(data, groups, **kwargs):
    """
    Visualize histogram of ungrouped observations. Typical usage examples:

    - Plotting a dataset:
        `plot_hist(data,group,**kw)`

    Parameters
    ==========
    groups : list, optional
        List of edge points in each interval. Each element can consist of a single or two edge values.
        Can either be a single list for all lists of observations or individual lists for each.
    """
    show = kwargs.get("show", True)
    backend = kwargs.get("backend", TWO_D_B)

    if backend != MB:
        raise NotImplemented

    kwargs.setdefault("ylabel", "Frekvensdensitet")

    f = frequency(data, groups, show=False)
    x, y, w = [], [], []
    for i, ff in enumerate(f.frekvens):
        width = f.observation[i][1] - f.observation[i][0]
        w.append(width)
        x.append((f.observation[i][1] + f.observation[i][0]) / 2)
        y.append(ff / width)

    series = BarplotSeries(x, y, x, width=w)
    series.rendering_kw.setdefault("linewidth", 3)
    p = backend(series, **kwargs)
    if show:
        p.show()
    return p


if __name__ == "__main__":
    pass
    # boxplot([1, 2, 3, 4, 5, 6, 6, 100])
    # a = boxplot([[1, 2, 3, 10], [1, 2, 3, 8]], labels=["hej", "dav"])
    # print(a)
    # boxplot([[1, 2, 3, 10], [1, 2, 3, 8]], groups=[1, 2, 3, 4, 5], labels=["test","test2"])

    # plot_bars([1, 2, 3, 3], labels=["hejsa"])
    # a = plot_bars([[1, 2, 3, 10, 10], [1, 2, 2, 3, 6, 8], [1, 2, 2, 3, 6, 8]], labels=["hejsa", "davs der", "3"])
    # print(a.series[0].get_data())

    # plot_sum([1, 2, 3, 3], labels=["hejsa"])
    # plot_sum([[1, 2, 3, 3], [2, 0, 4, 0]], groups=[1, 2, 3, 4, 5], labels=["hejsa", "test2"])
    # a = plot_sum([[1, 2, 3, 10, 10], [1, 2, 2, 3, 10], [1, 2, 2, 3, 6, 8]], labels=["hejsa", "davs der", "3"])

    # plot_hist([1, 3, 2], [1, 2, 3, 9])
